#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Main downloader module.

Implements:

* module function _ -- i18n translation wrapper.
* module function get_downloader -- return a new download controller
  object instance.
* module function i18n -- exportable alias for _(text) translation
  wrapper.
* module function shorten_path -- shorten file path to a limited number
  of columns.
* class DownloaderController -- main class for controlling a download
  procedure.
* class DownloaderError -- downloader error handler.
* class DownloaderMiningDude -- helper class for downloading procedure.
* class DownloaderModel -- main model class for downloading procedure.
* class DownloaderPriorityQueueTask -- task structure for PriorityQueue
  objects used during downloading procedure.
"""

"""
    The MIT License (MIT)

    Copyright (c) 2015+ Raphaël SEBAN <motus@laposte.net>

    Permission is hereby granted, free of charge, to any person
    obtaining a copy of this software and associated documentation
    files (the "Software"), to deal in the Software without
    restriction, including without limitation the rights to use, copy,
    modify, merge, publish, distribute, sublicense, and/or sell copies
    of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
    BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
    ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
    CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""

import os
import queue
import threading
import time
import urllib.parse
import urllib.request


def _ (text):
    """i18n translation wrapper.

    Parameter:
        text -- text string of chars to be translated (i18n support).

    This technique allows the use of internationalization (i18n)
    parsers such as GNU xgettext, by marking each string to be
    translated with _("text to translate").

    This also makes this module compatible with tkRAD i18n support,
    with nothing else than merely importing it and using it straight
    forward.
    """
    try:
        return __builtins__["_"](text)
    except:
        return text
    # end try
# end def


def get_downloader (**kw):
    """Return a new download controller object instance.

    Parameter:
        kw -- keyword arguments.

    Supported keyword arguments:
        url -- URL of remote file to download (default: None). This
        attribute MUST be set before running a download procedure.

        download_directory -- local path location of a target directory
        for downloading procedure (default: None).

        update_callback -- callback function or method to call at each
        downloading step (default: None). This callback is mainly used
        for information display purpose.

    Better use this helper function rather than implemented classes.
    """
    return DownloaderController(**kw)
# end def


def i18n (text):
    """Exportable alias for _(text) translation wrapper.

    Parameter:
        text -- text string of chars to be translated (i18n support).
    """
    return _(text)
# end def


def shorten_path (path, width=40, placeholder="..."):
    """Shorten file path to a limited number of columns.

    Parameters:
        path -- file path string of chars to shorten.

        width -- max. number of columns for resulting string of chars
        (default: 40).

        placeholder -- string of chars used to indicate path string has
        been actually shortened (default: '...'). Will raise a
        **ValueError** if an empty string of chars. Even when not used
        under certain circumstances, this parameter MUST be a plain
        string of chars.

    Returns a given file 'path' shortened to a string of chars of
    'width' maximum number of chars, including 'placeholder' chars,
    whenever necessary.
    """
    if not placeholder:
        raise ValueError(
            _("A placeholder string of chars is required.")
        )
    else:
        width = max(0, width)
        if len(path) > width:
            (_head, _tail) = os.path.split(path)
            placeholder = placeholder[:width]
            _swidth = width - len(placeholder)
            if _swidth > 0:
                _tail = _tail[-_swidth:]
                _swidth = max(0, _swidth - len(_tail))
                _head = _head[:_swidth]
            else:
                _head = _tail = ""
            # end if
            path = "{}{}{}".format(_head, placeholder, _tail)
        # end if
    # end if
    return path
# end def


class DownloaderController:
    """Main class for controlling a download procedure."""

    def __init__ (self, url=None, download_directory=None,
                                                update_callback=None):
        self.model = DownloaderModel(
            url=url,
            download_directory=download_directory,
            update_callback=update_callback,
        )
        self._started = False
    # end def

    def cancel_download (self, timeout=None):
        return self.model.cancel(timeout)
    # end def

    @property
    def download_directory (self):
        return self.model.download_directory
    # end def

    @download_directory.setter
    def download_directory (self, dir_path):
        self.model.download_directory = dir_path
    # end def

    def get_downloaded_filepath (self, timeout=None):
        return self.model.get_downloaded_filepath(timeout)
    # end def

    def is_downloading (self):
        return self.model.is_alive()
    # end def

    def start_download (self, url=None, download_directory=None,
                                                update_callback=None):
        if self.is_downloading():
            raise DownloaderError(_("Downloader is already running."))
            return
        elif self._started:
            self.model = DownloaderModel(
                url=self.url,
                download_directory=self.download_directory,
                update_callback=self.update_callback,
            )
        # end if
        self.url = url or self.url
        self.download_directory = (
            download_directory or self.download_directory
        )
        self.update_callback = update_callback or self.update_callback
        self.model.start()
        self._started = True
    # end def

    @property
    def update_callback (self):
        return self.model.update_callback
    # end def

    @update_callback.setter
    def update_callback (self, callback):
        self.model.update_callback = callback
    # end def

    @property
    def url (self):
        return self.model.url
    # end def

    @url.setter
    def url (self, url):
        self.model.url = url
    # end def

# end class


class DownloaderError (Exception):
    """Downloader error handler."""
    pass
# end class


class DownloaderMiningDude (threading.Thread):
    """Helper class for downloading procedure."""

    def __init__ (self, queue, update_callback):
        super().__init__()
        self.daemon = True
        self.queue = queue
        self.update_callback = update_callback
    # end def

    def run (self):
        _locked = False
        while not (_locked and self.queue.empty()):
            task = self.queue.get()
            data = task.data
            self.queue.task_done()
            if data is None:
                _locked = True
            elif not _locked:
                self.update_callback(**data)
            # end if
        # end while
    # end def

    @property
    def update_callback (self):
        return self.__update_callback
    # end def

    @update_callback.setter
    def update_callback (self, callback):
        if callable(callback):
            self.__update_callback = callback
        else:
            self.__update_callback = lambda *args, **kw: None
        # end if
    # end def

# end class


class DownloaderModel (threading.Thread):
    """Main model class for downloading procedure."""

    def __init__ (self, url=None, download_directory=None,
                                                update_callback=None):
        super().__init__()
        self.daemon = True
        self._running = False
        self._cancelled = False
        self._states = dict()
        self._block_size = 8192
        self._file_path = ""
        self._queue = queue.PriorityQueue()
        self.url = url
        self.download_directory = download_directory
        self.update_callback = update_callback
    # end def

    def _do_step_01 (self):
        self._file_path = ""
        _url = shorten_path(self.url)
        self._vars.update(short_url=_url)
        self._update_interface(
            url=self.url,
            shorten_url=_url,
            status="URL_OPEN",
            info=_("Trying to connect to server..."),
        )
        _attempt = self._vars.get("count", 0)
        if _attempt < 3:
            _response = urllib.request.urlopen(self.url)
            _size = _response.getheader("Content-Length")
            if _size:
                self._vars.update(
                    count=0,
                    response=_response,
                    content_length=int(_size),
                )
                return 2
            # end if
            self._vars.update(count=int(_attempt + 1))
            time.sleep(1.0)
            return 1
        else:
            raise DownloaderError(
                _("Unable to determine remote file size.")
            )
            self._running = False
            self._cancelled = True
            return
        # end if
    # end def

    def _do_step_02 (self):
        _ddir = self._get_available_directory(self.download_directory)
        _filename = self._get_available_filename(self.url, _ddir)
        _fpath = os.path.join(_ddir, _filename)
        _ftemp_path = _fpath + ".part"
        _fd = open(_ftemp_path, "wb")
        self._vars.update(
            available_dir=_ddir,
            available_filename=_filename,
            available_filepath=_fpath,
            available_fd=_fd,
            temporary_filepath=_ftemp_path,
        )
        return 3
    # end def

    def _do_step_03 (self):
        _count = self._vars.get("count", 0)
        _fd = self._vars.get("available_fd")
        _response = self._vars.get("response")
        _size = self._vars.get("content_length")
        _url = self._vars.get("short_url")
        self._info("DOWNLOAD", _("Downloading file..."))
        _data = _response.read(self._block_size)
        _fd.write(_data)
        _count += len(_data)
        self._vars.update(count=_count)
        self._progress(100 * _count // _size)
        if _count >= _size:
            _response.close()
            _fd.close()
            os.rename(
                self._vars.get("temporary_filepath"),
                self._vars.get("available_filepath")
            )
            self._file_path = self._vars.get("available_filepath")
            self._running = False
        # end if
        return 3 # loop on step 3
    # end def

    def _get_available_directory (self, dir_list):
        if dir_list:
            if isinstance(dir_list, str):
                dir_list = dir_list.split()
            # end if
            if isinstance(dir_list, (tuple, list)):
                for _dir in dir_list:
                    # ensure directory path is correct
                    _dir = os.path.abspath(
                        os.path.realpath(os.path.expanduser(_dir))
                    )
                    # directory must exist AND being writable
                    if os.path.isdir(_dir) and os.access(_dir, os.W_OK):
                        return _dir
                    # end if
                # end for
            else:
                raise DownloaderError(
                    _("Download directory must either be a "
                        "string of chars or a list of strings.")
                )
                return ""
            # end if
        # end if
        raise DownloaderError(_("No available download directory."))
        return ""
    # end def

    def _get_available_filename (self, url, target_dir):
        _fpath = urllib.parse.urlsplit(url).path
        (_head, _tail) = os.path.split(_fpath)
        (_prefix, _suffix) = os.path.splitext(_tail)
        _prefix = _prefix or "downloaded_file"
        _index = _sep = ""
        while True:
            _fname = "{}{}{}{}".format(_prefix, _sep, _index, _suffix)
            _fpath = os.path.join(target_dir, _fname)
            if not os.path.lexists(_fpath):
                return _fname
            # end if
            if not _index:
                _index = 0
                _sep = "-"
            # end if
            _index += 1
            if _index > 10000:
                break
            # end if
        # end while
        raise DownloaderError(_("No available filename."))
        return ""
    # end def

    def _info (self, state, message, priority=99):
        self._update_interface(
            status=str(state).upper(), info=message, priority=priority
        )
    # end def

    def _progress (self, value, priority=99):
        self._update_interface(progress=value, priority=priority)
    # end def

    def _purge_temporary_file (self):
        try:
            self._vars.get("available_fd").close()
        except:
            pass
        # end try
        try:
            os.remove(self._vars.get("temporary_filepath"))
        except:
            pass
        # end try
    # end def

    def _queue_put (self, data, priority=99):
        self._queue.put(DownloaderPriorityQueueTask(priority, data))
    # end def

    def _release_dude (self, priority=0):
        self._queue_put(None, priority)
    # end def

    def _update_interface (self, priority=99, **kw):
        self._states.update(kw)
        priority += int(kw.get("progress", 0))
        self._queue_put(self._states.copy(), priority)
    # end def

    def cancel (self, timeout=None):
        if self.is_alive():
            self._running = False
            self._cancelled = True
            self.join(timeout)
        # end if
        return not self.is_alive()
    # end def

    def get_downloaded_filepath (self, timeout=None):
        if self.is_alive():
            self.join(timeout)
        # end if
        return self._file_path
    # end def

    def run (self):
        mining_dude = DownloaderMiningDude(
            queue=self._queue, update_callback=self.update_callback
        )
        mining_dude.start()
        self._running = True
        self._cancelled = False
        self._vars = dict()
        _step = 1
        while self._running and not self._cancelled:
            _step = getattr(self, "_do_step_{:02d}".format(_step))()
        # end while
        self._purge_temporary_file()
        if self._cancelled:
            self._info("EXIT_CANCEL", _("Cancelled."), priority=1)
        else:
            self._info("EXIT_SUCCESS", _("Done."), priority=1)
        # end if
        self._release_dude(priority=2)
    # end def

# end class


class DownloaderPriorityQueueTask:
    """Task structure for PriorityQueue objects used during downloading
    procedure.
    """

    def __init__ (self, priority=99, data=None):
        self.priority = int(priority)
        self.data = data
    # end def

    def __lt__ (self, right_operand):
        try:
            return int(self.priority) < int(right_operand.priority)
        except:
            return NotImplemented
        # end try
    # end def

# end class
