#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Tkinter easy downloading tool module.

Implements:

* class EasyTool -- tkinter remote file downloading easy tool class.
* module function run_tool -- run tkinter downloading easy tool.
"""

"""
    The MIT License (MIT)

    Copyright (c) 2015+ Raphaël SEBAN <motus@laposte.net>

    Permission is hereby granted, free of charge, to any person
    obtaining a copy of this software and associated documentation
    files (the "Software"), to deal in the Software without
    restriction, including without limitation the rights to use, copy,
    modify, merge, publish, distribute, sublicense, and/or sell copies
    of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
    BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
    ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
    CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""

import tkinter as tk
from tkinter import ttk
from tkinter import filedialog
from tkinter import messagebox

try:
    from .tkinter_ui import *
except:
    from tkinter_ui import *
# end try


class EasyTool (tk.Tk):
    """Tkinter remote file downloading easy tool class."""

    def __init__ (self):
        super().__init__()
        self._setup_gui()
    # end def

    def _choose_directory (self):
        _dirpath = filedialog.askdirectory(
            title="Please, choose a download directory",
            initialdir=self.dnl_dir_cvar.get(),
            mustexist=True,
            parent=self,
        )
        if _dirpath:
            self.dnl_dir_cvar.set(_dirpath)
        # end if
    # end def

    def _download_from_dialog (self):
        self._status_reset()
        _file_path = tk_get_downloaded_filepath(
            url=self.url_cvar.get(),
            download_directory=self.dnl_dir_cvar.get(),
        )
        if _file_path:
            self._status_success(_file_path)
        else:
            self._status_cancel()
        # end if
    # end def

    def _go_download (self):
        if not (self.url_cvar.get() and self.dnl_dir_cvar.get()):
            messagebox.showwarning(
                title="Attention",
                message="Please, fill out URL and "
                    "download directory path.",
                parent=self,
            )
        else:
            self._download_from_dialog()
        # end if
    # end def

    def _on_quit_app (self, *args, **kw):
        self.destroy()
    # end def

    def _setup_gui (self):
        self.title("Easy tool")
        self.withdraw()
        self.resizable(width=False, height=False)
        self.protocol("WM_DELETE_WINDOW", self._on_quit_app)
        _frame = ttk.Frame(self, padding=5)
        ttk.Label(_frame, text="File to download:").pack(fill=tk.X)
        ttk.Label(_frame, text="URL: ").pack(side=tk.LEFT)
        self.url_cvar = tk.StringVar()
        ttk.Entry(
            _frame, textvariable=self.url_cvar, width=40,
        ).pack(side=tk.RIGHT, expand=1, fill=tk.X)
        _frame.pack(expand=1, fill=tk.X)
        _frame = ttk.Frame(self, padding=5)
        ttk.Label(_frame, text="Download directory:").pack(fill=tk.X)
        self.dnl_dir_cvar = tk.StringVar()
        self.dnl_dir_cvar.set("~")
        ttk.Entry(
            _frame, textvariable=self.dnl_dir_cvar
        ).pack(side=tk.LEFT, expand=1, fill=tk.X)
        ttk.Label(_frame, text=" ").pack(side=tk.LEFT)
        ttk.Button(
            _frame,
            text="Choose directory",
            command=self._choose_directory,
        ).pack(side=tk.RIGHT)
        _frame.pack(expand=1, fill=tk.X)
        _frame = ttk.Frame(self, padding=5)
        ttk.Label(_frame, text="Status: ").pack(side=tk.LEFT)
        self._status_cvar = tk.StringVar()
        ttk.Label(
            _frame, textvariable=self._status_cvar, foreground="red",
        ).pack(side=tk.RIGHT, expand=1, fill=tk.X)
        _frame.pack(anchor=tk.W)
        _frame = ttk.Frame(self, padding=5)
        ttk.Button(
            _frame, text="Download", command=self._go_download,
        ).pack(side=tk.LEFT, padx=10)
        ttk.Button(
            _frame, text="Quit", command=self._on_quit_app,
        ).pack(side=tk.RIGHT)
        _frame.pack()
    # end def

    def _status_cancel (self, reason=""):
        self._status_cvar.set(
            "\n".join(("download cancelled.", str(reason))).strip()
        )
    # end def

    def _status_reset (self):
        self._status_cvar.set("")
    # end def

    def _status_success (self, file_path):
        self._status_cvar.set(
            "\n".join((
                "downloaded file is available at:", str(file_path)
            )).strip()
        )
    # end def

    def run (self):
        """Run tkinter downloading easy tool."""
        self.deiconify()
        self.mainloop()
    # end def

# end class


def run_tool ():
    """Run tkinter downloading easy tool."""
    EasyTool().run()
# end def


if __name__ == "__main__":
    run_tool()
# end if
