#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Tkinter UI downloader adaptation module.

Implements:

* module function run_demo -- run tkinter demonstration procedure.
* module function tk_get_downloaded_filepath -- tkinter helper for easy
  download of remote file.
* class tkDownloaderDialog -- tkinter generic download display dialog.
* class ttkDownloaderFrame -- tkinter.ttk downloader graphical frame
  component.
"""

"""
    The MIT License (MIT)

    Copyright (c) 2015+ Raphaël SEBAN <motus@laposte.net>

    Permission is hereby granted, free of charge, to any person
    obtaining a copy of this software and associated documentation
    files (the "Software"), to deal in the Software without
    restriction, including without limitation the rights to use, copy,
    modify, merge, publish, distribute, sublicense, and/or sell copies
    of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
    BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
    ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
    CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""

import tkinter as tk
from tkinter import ttk

try:
    from .downloader import *
except:
    from downloader import *
# end try

# import alias for translation wrapper
_ = i18n


def tk_get_downloaded_filepath (url, download_directory):
    """Tkinter helper for easy download of remote file.

    Parameters:
        url -- URL of remote file to download.

        download_directory -- local path location of a target directory
        for downloading procedure.

    This helper function shows a tkinter information dialog for
    downloading procedure, blocks calling tkinter parent window until
    download is complete (or cancelled), then returns file path of
    downloaded local file (or empty string on cancellation).
    """
    try:
        tk_parent = tk._default_root.winfo_toplevel()
    except:
        tk_parent = None
    # end try
    return tkDownloaderDialog(tk_parent).get_downloaded_filepath(
        url, download_directory
    )
# end def


class tkDownloaderDialog (tk.Toplevel):
    """Tkinter generic download display dialog."""

    def __init__ (self, master=None, **options):
        super().__init__(master, **options)
        self.__downloaded_filepath = ""
        self._setup_widget()
    # end def

    def _center_window (self):
        self.update_idletasks()
        if self.master:
            _root = self.master.winfo_toplevel()
            _rw, _rh = _root.winfo_width(), _root.winfo_height()
            _x0, _y0 = _root.winfo_x(), _root.winfo_y()
        else:
            _rw = self.winfo_screenwidth()
            _rh = self.winfo_screenheight()
            _x0 = _y0 = 0
        # end if
        _w, _h = self.winfo_reqwidth(), self.winfo_reqheight()
        self.geometry(
            "+{x}+{y}".format(x=_x0 + (_rw-_w)//2, y=_y0 + (_rh-_h)//2)
        )
    # end def

    def _close_dialog (self):
        self.update_idletasks()
        self.after(500, self.destroy)
    # end def

    def _on_download_cancel (self, *args, **kw):
        self.__downloaded_filepath = ""
        self._close_dialog()
    # end def

    def _on_download_success (self, file_path, *args, **kw):
        self.__downloaded_filepath = file_path
        self._close_dialog()
    # end def

    def _setup_widget (self):
        self.withdraw()
        self.transient(self.master)
        self.resizable(width=False, height=False)
        self.frame = ttkDownloaderFrame(self, padding=10)
        self.frame.pack()
        self.protocol("WM_DELETE_WINDOW", self.on_quit_dialog)
    # end def

    def download (self, url, download_directory):
        self.frame.download(
            url=url,
            download_directory=download_directory,
            on_success=self._on_download_success,
            on_cancel=self._on_download_cancel,
        )
    # end def

    def get_downloaded_filepath (self, url, download_directory):
        self.download(url, download_directory)
        self.show()
        self.grab_set()
        self.wait_window(self)
        return self.__downloaded_filepath
    # end def

    def on_quit_dialog (self, *args, **kw):
        self.frame._on_download_cancel()
    # end def

    def show (self, *args, **kw):
        self.deiconify()
        self._center_window()
    # end def

# end class


class ttkDownloaderFrame (ttk.Frame):
    """Tkinter.ttk downloader graphical frame component.

    This graphical component may be embedded into a dialog window or in
    any other tkinter widget container.
    """

    def __init__ (self, master=None, downloader=None, **options):
        super().__init__(master, **options)
        self.downloader = downloader or get_downloader()
        self.downloader.update_callback = self._update_states
        self.on_success_callback = None
        self.on_cancel_callback = None
        self._setup_widget()
    # end def

    def _on_download_cancel (self, event=None):
        self.cancel_download()
        self.update_idletasks()
        if callable(self.on_cancel_callback):
            self.after_idle(self.on_cancel_callback)
        # end if
    # end def

    def _on_download_success (self, event=None):
        self.update_idletasks()
        if callable(self.on_success_callback):
            self.after_idle(
                lambda:
                self.on_success_callback(
                    file_path=self.downloader.get_downloaded_filepath(),
                )
            )
        # end if
    # end def

    def _setup_widget (self):
        self.label_cvar = tk.StringVar()
        self.label = ttk.Label(self, textvariable=self.label_cvar)
        self.label.pack(expand=1, fill=tk.X, pady=5)
        self.progressbar_cvar = tk.DoubleVar()
        self.progressbar = ttk.Progressbar(
            self, orient=tk.HORIZONTAL, variable=self.progressbar_cvar
        )
        self.progressbar.pack(side=tk.LEFT, expand=1, fill=tk.X, pady=5)
        self.button = ttk.Button(
            self, text=_("Cancel"), command=self._on_download_cancel
        )
        self.button.pack(side=tk.RIGHT, padx=5, pady=5)
    # end def

    def _start_download (self, url=None, download_directory=None):
        self.downloader.start_download(
            url=url, download_directory=download_directory
        )
    # end def

    def _update_states (self, status="", info="", progress=0, **kw):
        self.label_cvar.set(
            "\n".join((
                str(info),
                _("URL: {url}").format(url=kw.get("shorten_url") or "")
            )).strip()
        )
        self.progressbar_cvar.set(progress)
        if status == "URL_OPEN":
            self.progressbar.configure(mode="indeterminate")
            self.progressbar.start()
        elif status == "DOWNLOAD":
            self.progressbar.configure(mode="determinate")
            self.progressbar.stop()
        elif status == "EXIT_SUCCESS":
            # better use scheduled task to release ending thread
            self.update_idletasks()
            self.after_idle(self._on_download_success)
        # end if
    # end def

    def cancel_download (self, timeout=None):
        self.update_idletasks()
        self.downloader.cancel_download(timeout)
    # end def

    def download (self, url, download_directory, on_success=None,
                                                        on_cancel=None):
        self.on_success_callback = on_success
        self.on_cancel_callback = on_cancel
        self._start_download(url, download_directory)
    # end def

    def reset (self, *args, **kw):
        self._update_states()
    # end def

# end class


def run_demo ():
    """Run tkinter demonstration procedure."""
    class TkinterDemo (tk.Tk):

        URL = (
            "https://github.com/tarball69/tkBoulderDash/archive/master.zip"
        )

        DOWNLOAD_DIRS = (
            "~/tmp ~/temp ~/download ~/Download "
            "~/downloads ~/Downloads ~ /tmp"
        )

        def __init__ (self):
            super().__init__()
            self.setup_gui()
        # end def

        def download_from_dialog (self):
            self.status_reset()
            _file_path = tk_get_downloaded_filepath(
                self.URL, self.DOWNLOAD_DIRS
            )
            if _file_path:
                self.status_success(_file_path)
            else:
                self.status_cancel()
            # end if
        # end def

        def download_in_frame (self):
            self.status_reset()
            self.dnlframe.download(
                self.URL, self.DOWNLOAD_DIRS,
                on_success=self.on_download_success,
                on_cancel=self.on_download_cancel,
            )
        # end def

        def on_download_cancel (self, *args, **kw):
            self.status_cancel()
        # end def

        def on_download_success (self, file_path, *args, **kw):
            self.status_success(file_path)
        # end def

        def on_quit_demo (self, *args, **kw):
            self.dnlframe.cancel_download()
            self.destroy()
        # end def

        def run (self):
            self.deiconify()
            self.mainloop()
        # end def

        def setup_gui (self):
            self.title(_("Tkinter demo"))
            self.withdraw()
            self.resizable(width=False, height=False)
            self.protocol("WM_DELETE_WINDOW", self.on_quit_demo)
            _frame = ttk.Frame(self, padding=5)
            ttk.Label(
                _frame, text=_("Click the following button to ")
            ).pack(side=tk.LEFT)
            ttk.Button(
                _frame,
                text=_("download in frame"),
                command=self.download_in_frame,
            ).pack(side=tk.RIGHT)
            _frame.pack(anchor=tk.W)
            self.dnlframe = ttkDownloaderFrame(
                self, downloader=None, padding=5
            )
            self.dnlframe.pack(expand=1, fill=tk.BOTH)
            _frame = ttk.Frame(self, padding=5)
            ttk.Label(
                _frame, text=_("Click the following button to ")
            ).pack(side=tk.LEFT)
            ttk.Button(
                _frame,
                text=_("download from a dialog box"),
                command=self.download_from_dialog,
            ).pack(side=tk.RIGHT)
            _frame.pack(anchor=tk.W)
            _frame = ttk.Frame(self, padding=5)
            ttk.Label(_frame, text=_("Status: ")).pack(side=tk.LEFT)
            self.status_cvar = tk.StringVar()
            ttk.Label(
                _frame, textvariable=self.status_cvar, foreground="red",
            ).pack(side=tk.RIGHT, expand=1, fill=tk.X)
            _frame.pack(anchor=tk.W)
            _frame = ttk.Frame(self, padding=5)
            ttk.Button(
                _frame, text=_("Quit demo"), command=self.on_quit_demo,
            ).pack()
            _frame.pack()
        # end def

        def status_cancel (self, reason=""):
            self.status_cvar.set(
                "\n".join((_("download cancelled."), str(reason))).strip()
            )
        # end def

        def status_reset (self):
            self.status_cvar.set("")
        # end def

        def status_success (self, file_path):
            self.status_cvar.set(
                "\n".join((
                    _("downloaded file is available at:"), str(file_path)
                )).strip()
            )
        # end def

    # end class

    TkinterDemo().run()
# end def


if __name__ == "__main__":
    run_demo()
# end if
